;;; -*- auto-recompile: t -*-

;;; csound-eel.el --- alternate font-lock settings
;;;                    and extra stuff for csound-orc-mode

;; Keywords: csound

;; This file is not part of GNU Emacs.
;; 
;; csound-eel.el is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; csound-eel.el is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Author: Stphane Rollandin <hepta@zogotounga.net>
;;---------

;;; Commentary:
;;-------------

;;    full documentation here:
;;    http://www.zogotounga.net/comp/csoundx.html

;;
;;; Installation:
;;                  
;; ==========================================================
;; this file should be installed through the csound-x package
;; ==========================================================
;;

;; last modified January 6, 2004

;;; Code:

(require 'embedded-elisp-library)

;; ===================================================================
;;                          initialization
;; ===================================================================

;; defining new allowed buttons for <ELISP> areas:
(add-to-list 'embedded-buttons-alist '("export" . 'cseel-export-this-snapshot))
(add-to-list 'embedded-buttons-alist '("restore" . 'eval-this-embedded-elisp))
(add-to-list 'embedded-buttons-alist '("see/hide" . 'collapse-unfold-this-embedded-elisp))

;; and taking them into account immediately:
(eel-wake-up-embedded-buttons t)


;; ===================================================================
;;          misc. utilities (library for embedded elisp)
;; ===================================================================

(defun cseel-new-section ()
  "Go to the end of the score, insert a s statement"
  (cscsd-goto-end-of-sco)
  (unless (re-search-backward "^[ \t]*e[ \t]*$" nil t)
    (search-backward "</CsScore>" nil t)
    (save-excursion
      (insert "e\n")))
  (insert "\ns\n\n")
  (forward-line -1))


;; ===================================================================
;;                        general-purpose macros
;; ===================================================================


(defmacro csd-process-variation (&rest body)
  "Copy the current buffer in a temporary buffer, evaluate BODY there,
write the buffer in `cscsd-temp-csd-file' and process it with `cscsd-process'"
  `(let ((this-buffer (current-buffer)))
     (with-temp-file cscsd-temp-csd-file
       (insert-buffer this-buffer)
       ,@body)
     (cscsd-process cscsd-temp-csd-file)))
	   	   
(defmacro csd-process-variation-as (name &rest body)
  "Copy the current buffer in a temporary buffer, evaluate BODY there,
write the buffer in `cscsd-temp-csd-file' and process it with `cscsd-process'
the audio output is called NAME"
  `(let ((this-buffer (current-buffer)))
     (with-temp-file cscsd-temp-csd-file
       (insert-buffer this-buffer)
       ,@body)
     (cscsd-process cscsd-temp-csd-file ,name)))
     
	   	   

;; ===================================================================
;;                           meta tags queries
;; ===================================================================

(defvar cseel-meta-tag ";|"
  "sco/orc comment beginning as such will be recognized as meta instructions")

(defun cseel-goto-tag (tag &rest tags)
  "Move to the first occurrence of `cseel-meta-tag' followed by TAGS
all TAGS should be words, with no spaces"
  (goto-char (point-min))
  (re-search-forward (concat cseel-meta-tag 
			     tag
			     (mapconcat(lambda (s)
				       (concat "[ \t]+" s)) tags "")  
			     "[ \t]*$") nil t))

(defun cseel-get-next-tag (&optional tag)
  "Move to the next occurrence of `cseel-meta-tag'
returns the associated keywords, as a list"
  (if tag
      (let (tags ok)
	(while (and (re-search-forward (concat cseel-meta-tag "\\(.*\\)") nil t)
		    (null (setq ok (string= 
				    (car (setq tags (split-string (match-string-no-properties 1))))
				    tag)))))
	(if ok tags nil))
    (if (re-search-forward (concat cseel-meta-tag "\\(.*\\)") nil t)
	(split-string (match-string-no-properties 1))))) 
	
	     
;; ===================================================================
;;                          ;|begPublic  
;;                          ;|endPublic      
;; ===================================================================
 
(defun cseel-display-all-public-sections ()
  (when cscsd-use-MMM
    (cscsd-mmm))
  (save-excursion
    (goto-char (point-min))
    (let (name)
      (while (setq name (cseel-get-next-tag "begPublic"))
	(let* ((beg (point))
	       (end (progn (cseel-get-next-tag "endPublic")
			   (re-search-backward cseel-meta-tag)
			   (point)))
	       (olay (find-if (lambda (o) (overlay-get o 'mmm))
			      (overlays-in beg end))))
	(save-excursion
	  (set-buffer 
	   (make-indirect-buffer (current-buffer) 
				 (format "%s in CSD %s" (or (cadr name) "?") (buffer-name))
				 t)) ;nil)) ;t))
	  (narrow-to-region beg end)
	  (new-frame (list (cons 'name (or (cadr name)
					   (format "CSD %s" (buffer-name))))
			   (cons 'width (or (and (caddr name) (read (caddr name)))
					    (frame-parameter (selected-frame) 'width)))
			   (cons 'height (or (and (cadddr name) (read (cadddr name)))
					     (min (+ 3 (count-lines beg end))
						  (frame-parameter (selected-frame) 'height))))))
	  ;; restoring MMM submodes:
	  (when (and olay cscsd-use-MMM (overlay-get olay 'submode))
	    (let ((submode (overlay-get olay 'submode)))
	      (mmm-clear-overlays beg end)
	      (mmm-make-region submode (point-min) (point-max)))))))
      ;; no effect ??
      (cscsd-build-menus))))

(defvar cseel-csd-frame-parameters ()
  "")

(defun cseel-display-only-public-sections ()
  (interactive)
  (let ((f (selected-frame)))
    (cseel-display-all-public-sections)
    (setq cseel-csd-frame-parameters
	  (plist-put cseel-csd-frame-parameters
		     (buffer-name)
 		     (delete-if (lambda (pair)
 				  (null (member (car pair) '(top left width height name title))))
 				(frame-parameters f))))
    (delete-frame f)
    (cscsd-build-menus)))

(defun cseel-collapse-public-sections ()
  (interactive)
  (switch-to-buffer (buffer-base-buffer))
  (cseel-delete-public-frames)
  (set-buffer-modified-p t)
  (modify-frame-parameters (selected-frame) (plist-get cseel-csd-frame-parameters (buffer-name)))
  (cscsd-build-menus)
  (when cscsd-use-MMM
    (cscsd-mmm)))

(defun cseel-delete-public-frames ()
  (let ((reg (concat " in CSD " (car (split-string (buffer-name) "\\." t)))))
    (dolist (f (frame-list))
      (when (and (string-match reg (buffer-name (window-buffer (frame-first-window f))))
		 (> (length (frame-list)) 1))
	(delete-frame f)))
    (dolist (b (buffer-list))
      (when (string-match reg (buffer-name b))
	(save-excursion
	  (set-buffer b)
	  (set-buffer-modified-p nil)
	  (kill-buffer b))))))


;; ===================================================================
;;          ;|menu           ;|toggles
;;          ;|endmenu        ;|endtoggles  
;; ===================================================================
 
(defmacro with-cs-syntax (&rest body)
  "Attempt to have a correct comment syntax anywhere in a MMM-enabled csd file"
  `(with-syntax-table csound-sco-mode-syntax-table
     (let ((comment-start ";") (comment-end ""))
       ,@body)))

(defun cseel-uncomment-lines-below ()
  "Uncomment all lines between points and the next meta tag"
  (let ((beg (point-at-bol 2))
	(end (save-excursion (cseel-get-next-tag)
			     (point-at-bol))))
    (when (> (- end beg) 1)
      (with-cs-syntax (uncomment-region beg end)))))

(defun cseel-comment-lines-below ()
  "Comment all lines between points and the next meta tag"
  (unless (save-excursion (forward-line 1) (looking-at "^;"))
    (let ((beg (point-at-bol 2))
	  (end (save-excursion (cseel-get-next-tag)
			       (point-at-bol))))
      (when (> (- end beg) 1)
	(with-cs-syntax (comment-region beg end))))))

(defun cseel-menu-item-active-p (name item)
  "Check weither ITEM in meta menu/toggles NAME is currently uncommented"
  (save-excursion
    (or (cseel-goto-tag "menu" name)
	(cseel-goto-tag "toggles" name))
    (if (cseel-get-next-tag item)
	(null (save-excursion (forward-line 1) (looking-at "^;")))
      nil)))

(defun cseel-toggle-menu-item (name item)
  "Toggle commented status of ITEM in meta toggles NAME"
  (interactive)
  (save-excursion
    (cseel-goto-tag "toggles" name)
    (cseel-get-next-tag item)
    (if (cseel-menu-item-active-p name item)    
	(cseel-comment-lines-below)
      (cseel-uncomment-lines-below))))

(defun cseel-set-menu-item (name &optional item)
  "Comment out all items in meta menu/toggles NAME, then uncomment ITEM if provided"
  (interactive)
  (save-excursion
    (or (cseel-goto-tag "menu" name)
	(cseel-goto-tag "toggles" name))
    (while (null (member (car (cseel-get-next-tag)) '("endmenu" "endtoggles")))
      (cseel-comment-lines-below))
    (when item
    (or (cseel-goto-tag "menu" name)
	(cseel-goto-tag "toggles" name))
      (if (cseel-get-next-tag item)
	  (cseel-uncomment-lines-below)))))


;; ===================================================================
;;                         "Csd" submenu
;; ===================================================================

(defmacro cseel-menu-collect-while (test &rest body)
  "collect a list of vectors for menu item specifications
BODY must evaluate to such a vector, or possibly to a submenu spec
BODY may also return nil, in which case no item is created
TEST is performed as long as it is true, the macro starting with point in \(point-min)"
  `(save-excursion
     (goto-char (point-min))
     (let (entries this-entry)
       (while ,test
	 (setq this-entry (progn ,@body))
	 (when this-entry
	   (add-to-list 'entries this-entry t)))
       entries)))

(put 'cseel-menu-collect-while 'lisp-indent-function 1)

(defun cseel-menu ()
  (delete-if 'null
   `(("Meta-Settings"
     ["Take a snapshot" cseel-store-snapshot (cscsd-non-indirect-buffer-p)]
     ;; ----------------------------  snapshots
     ("Restore snapshot"
      ,@(cseel-menu-collect-while (re-search-forward "<ELISP>.*SNAPSHOT: \\([^ \t\n]+\\)" nil t)
	  (let ((name (match-string-no-properties 1)))
	    (vector name
		    `(progn (goto-char (point-min))
			    (if (re-search-forward (concat "<ELISP>.*SNAPSHOT: " ,name) nil t)
				(eval-this-embedded-elisp)))
		    t)))
     )
     "--"  
     ;; ----------------------------  ;|menu  &  ;|toggles
     ,@(let (tags)
       (cseel-menu-collect-while (setq tags (cseel-get-next-tag))
	   (if (or (string= (car tags) "menu")
		   (string= (car tags) "toggles"))
	       (let* ((name (cadr tags))
		      end-menu action item
		      (menu (list name)))
		 (save-excursion
		   (if (cseel-goto-tag "menu" name) (setq end-menu "endmenu"
							  action 'cseel-set-menu-item)
		     (if (cseel-goto-tag "toggles" name) (setq end-menu "endtoggles"
							       action 'cseel-toggle-menu-item)))
		   (while (null (string= (setq item (car (cseel-get-next-tag))) end-menu))
		     (add-to-list 'menu (vector item (list action name item)
						:style 'toggle
						:selected `(cseel-menu-item-active-p ,name ,item))
				  t))
		   menu)))))
     "--"
     ;; ----------------------------  #define
     ,@(cseel-menu-collect-while (re-search-forward "^[ \t]*#define[ \t]*\\([^ \t#\n\r]+\\)" nil t)
	 (let ((name (match-string-no-properties 1) )
	      (section (cscsd-current-section)))
	  (vector (delete ?\n (concat "(" (if (eq section 'orc) "orc)" "sco)") " #define "
				      (car (split-string name  "(" t)))) 
		  `(progn ,(if (eq section 'orc) '(cscsd-goto-orc) '(cscsd-goto-sco))
			  (re-search-forward ,(concat "^[ \t]*#define[ \t]*" name "[ \t]*")))
		  t))))
     ;; ----------------------------  ;|begPublic   & ;|endPublic 
    ,(if (and (buffer-base-buffer)
	       (string-match ".* in CSD .*" (buffer-name)))
	["Full editing mode" cseel-collapse-public-sections t]
       (when (and (not (buffer-base-buffer))
		  (save-excursion (cseel-get-next-tag "begPublic")))
	 ["Public mode" cseel-display-only-public-sections t])))))
   

;; ===================================================================
;;                       snapshots
;; ===================================================================

(defun cseel-get-snapshot ()
  "Return the snapshot data for the current state of the csd"
  (save-excursion
    (let (tags in-tags snapshot)
      ;; #defines
      (goto-char (point-min))
      (while (re-search-forward
	      "^[ \t]*#define[ \t]+\\([a-zA-Z0-9]+\\)[ \t]+#\\([^#]+\\)#"
	      nil t)
	(setq snapshot (append snapshot (list (list :define (match-string-no-properties 1)
						    :value (match-string-no-properties 2)
						    :in (cscsd-current-section))))))
      ;; meta menus/ktoggles
      (goto-char (point-min))
      (while (setq tags (cseel-get-next-tag))
	(if (or (string= (car tags) "menu")
		(string= (car tags) "toggles"))
	    (while (and (setq in-tags (cseel-get-next-tag))
			(null (or (string= (car in-tags) "endmenu")
				  (string= (car in-tags) "endtoggles"))))
	      (setq snapshot
		    (append snapshot
			    (list
			     (list :menu (cadr tags) 
				   :type (car tags)
				   :item (car in-tags)
				   :active (null (save-excursion
						   (forward-line 1) 
						   (looking-at "^;"))))))))))
      snapshot)))

(defun cseel-restore-snapshot (snapshot)    
  "Restore the snapshot at point
this function must be invoked while point is on the corresponding <ELISP> line"
  (when (yes-or-no-p "restore this snaphot and forget current settings ?")
    (dolist (s snapshot)
      (when (plist-get s :menu)                    ;; idiot: c'est rpt pour chaque item...
	(cseel-set-menu-item (plist-get s :menu))))
    (dolist (s snapshot)
      (when (and (plist-get s :menu)
		 (plist-get s :active))
	(cseel-set-menu-item (plist-get s :menu) (plist-get s :item)))
      (when (plist-get s :define)
	(cscsd-set-macro-def (plist-get s :define) (plist-get s :value) (plist-get s :in))))))

(defun cseel-store-snapshot (&optional comment)
  "Write the <ELISP> area allowing the restoration of the current settings
COMMENT should be one word, the snapshot name"
  (interactive)
  (setq comment (or comment (read-from-minibuffer "Name: ")))
  (save-excursion
    (goto-char (point-min))
    (insert "<ELISP>")
    (save-excursion
      (insert " ;SNAPSHOT: " comment " [export] [restore] [see/hide]\n"
	      "(require 'csound-eel)\n(cseel-restore-snapshot '"
	      (with-output-to-string (prin1 (cseel-get-snapshot)))
	      ")"
	      "\n</ELISP>\n\n")
      (if (and cscsd-use-MMM mmm-mode) (cscsd-turn-mmm-on))
      (eel-wake-up-embedded-buttons))
    (collapse-unfold-this-embedded-elisp)))

(defun cseel-export-this-snapshot (&optional filename)   ;  terminer
  "Write a csd files with the snapshot at point restored
removes all elisp areas and meta instruction from this file
the current buffer is not changed"
  (interactive)
  (unless (string= (setq filename (or filename (read-file-name 
						"Export to csd file: "
						(file-name-directory (buffer-file-name)))))
		   "")
    (let ((buff (current-buffer))
	  (pos (point)))
      (with-temp-file filename
	(insert-buffer buff)
	(goto-char pos)
	(eval-this-embedded-elisp)
	;; (eel-remove-all-elisp)      ; ...  debugger, celui-ci
	;; (cscsd-expand-all-defines)     ; devrait simplifier les #define ...
	;; (cseel-remove-all-metatags)    ; devrait simplifier les mta instructions
	))))


;; === this is it.
(provide 'csound-eel)

;; csound-eel.el ends here

