;;; -*- auto-recompile: t -*-

;; This file is not part of GNU Emacs.
;; 
;; erman.el is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; erman.el is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Commentary:
;;-------------
;;             interacting with Surmulot manager via TCP/IP

;; last modified September 21, 2004
;; for Emacs 21.1
;;; --------------------------------------------------------------------------------

(require 'cl)

;;; Code:

(defgroup surmulot-manager nil
  "Inserting Emacs within muO local communication network"
  :group 'processes
  :prefix "erman-")

(defcustom erman-port 6666
  "Default TCP/IP port in local host. 
Erlang note:
this is the value ?EmacsPort which is defined in file
musicnet.erl within `erman-code-directory'"
  :type 'integer
  :group 'surmulot-manager)

(defcustom erman-repl-sit-for 10
  "How long should we display evaluation results in the minibuffer
within a REPL session. see commands `keykit-repl' and `squeak-repl'"
  :type 'integer
  :group 'surmulot-manager)

(defcustom erman-is-daemon nil
  "When t we reconnect upon deconnection"
  :type 'boolean
  :group 'surmulot-manager)

(defvar erman-process-timer nil
  "the timer enabling auto-reconnection")

(defvar erman-process nil
  "the manager interaction TCP/IP process")

(defvar erman-separator "\v"
  "one-character string used to separate items in a TCP request")

(defvar erman-end-separator "\v\v"
  "string used to separate TCP requests")

(defvar erman-last-answer nil
  "last answer from manager to a query from emacs")

(defvar erman-pending-input ""
  "accumulated incoming strings from manager.
This is parsed in erman-process-filter")

;;; --------------------------------------------------------------------------------
;;;
;;; connection to manager
;;;
;;; --------------------------------------------------------------------------------

; (erman-connect)
(defun erman-connect (&optional ntry wait-for)
  "Set up the connection to the manager TCP/IP server"
  (interactive)
  (unless (erman-connected-p)
    (setq ntry (or ntry 1)
          wait-for (or wait-for 3))
    (while (not (zerop ntry))
      (condition-case nil
          (when (setq erman-process (open-network-stream "erman" nil "127.0.0.1" erman-port))
            (erman-cancel-daemon)
            (setq ntry 0)
            (set-process-sentinel erman-process 'erman-process-sentinel)
            (set-process-filter erman-process 'erman-process-filter))
        (error (unless (zerop (setq ntry (1- ntry)))      
                 (sit-for wait-for)))))))

(defun erman-process-sentinel (process event)
  (message "manager: %s" event)
  (when erman-is-daemon 
    (erman-start-daemon)))

(defun erman-start-daemon ()
  (interactive)
  (erman-cancel-daemon)
  (setq erman-process-timer (run-at-time t 5 'erman-connect)))

(defun erman-cancel-daemon ()
  (interactive)
  (when (timerp erman-process-timer)
    (cancel-timer erman-process-timer)
    (setq erman-process-timer nil)))

(defun erman-daemon-waiting-p ()
  (timerp erman-process-timer))

(defun erman-process-filter (process strinput)
  (setq erman-pending-input (concat erman-pending-input strinput))
  (when (string-match erman-end-separator erman-pending-input)
    (while (string-match (concat "\\(.+?" erman-end-separator "\\)\\(\\(.\\|\n\\)*\\)")
			 erman-pending-input)
      (let* ((rest (or (match-string 2 erman-pending-input) ""))
	     (command (split-string (match-string 1 erman-pending-input) erman-separator))
	     (data (first command))
	     (manager (second command))
	     (type (third command))
	     (tid (fourth command)))
        (setq erman-pending-input rest)
	(when (string= type "Q")
          (process-send-string erman-process
                               (concat (condition-case nil
                                           (prin1-to-string (eval (read data)))
                                         (error "evaluation error")) erman-separator
                                         manager erman-separator 
                                         "A" erman-separator
                                         tid erman-end-separator)))
	(when (string= type "A")
	  (setq erman-last-answer data))))))

(defun erman-disconnect ()
  (interactive)
  (delete-process erman-process))

(defun erman-connected-p ()
  (interactive)
  "return nil if Emacs is not connected to its manager"
  (eq (process-status "erman") 'open))


;;; --------------------------------------------------------------------------------
;;;
;;; interaction with the managers
;;;
;;; --------------------------------------------------------------------------------

(defun erman-manager-connected-p (manager)
  (and (erman-connected-p )
       (string= (erman-ask "super_listener" manager) "true")))

(defun erman-squeak-connected-p ()
  (interactive)
  (erman-manager-connected-p "squeak_manager"))

(defun erman-keykit-connected-p ()
  (interactive)
  (erman-manager-connected-p "keykit_manager"))

;TEST (erman-ask "super_listener" "ping ?") => "42"
;TEST (erman-ask "keykit_manager" "fractal('a,e,f')") => "'ad32,e,f,e,bo2,co3,f,c,c+'"
;TEST (erman-ask "squeak_manager" "5 sin") => "-0.958924274663138"
;TEST (length (erman-ask "keykit_manager" "fractal(fractal(fractal('a,e,f')))")) => 18118

(defun erman-ask (manager question)
  ""
  (when (erman-connected-p)
    (setq erman-last-answer nil)
    (process-send-string erman-process (concat question erman-separator
                                               manager erman-separator
                                               "Q" erman-separator
                                               "0" erman-end-separator))
    (while (null erman-last-answer)
      (accept-process-output erman-process))
    erman-last-answer))

(defun erman-ask-squeak (question)
  (interactive "sSqueak> ")
  (let ((ans (erman-ask "squeak_manager" question)))
    (if (interactive-p) (message ans) ans)))

(defun erman-ask-keykit (question)
  (interactive "sKeykit> ")
  (let ((ans (erman-ask "keykit_manager" question)))
    (if (interactive-p) (message ans) ans)))

(defun erman-repl (function)
  (while t
    (call-interactively function)
    (sit-for erman-repl-sit-for)))

(defun squeak-repl ()
  (interactive)
  (erman-repl 'erman-ask-squeak))

(defun keykit-repl ()
  (interactive)
  (erman-repl 'erman-ask-keykit))


;;; --------------------------------------------------------------------------------
;;;
;;; Erlang implementation for Surmulot manager
;;;
;;; --------------------------------------------------------------------------------

(require 'erlang-start nil t)

(defcustom erman-Erlang-code-directory "~/site-lisp/stef-elisp/erlang"
  "Directory where lives the Erlang code for managers"
  :type 'string
  :group 'surmulot-manager)

(defcustom erman-Erlang-start-command "musicnet:start_servers()."
  "Erlang command for starting the managers servers"
  :type 'string
  :group 'surmulot-manager)

(defun erman-start-Erlang-managers ()
  "start Erlang, set up all manager servers, then connect Emacs to its manager"
  (interactive)
  (require 'erlang)
  (let ((curdir default-directory))
    (save-excursion
      (unless (inferior-erlang-running-p)
	(cd erman-Erlang-code-directory)
	(erlang-shell)
	(inferior-erlang-wait-prompt)
	(inferior-erlang-send-command erman-Erlang-start-command)
	(inferior-erlang-wait-prompt)
	(erman-connect)))
    (cd curdir)
    (switch-to-buffer (current-buffer))))


;;; --------------------------- End  -----------------------------

(provide 'erman)

;;; erman ends here

